@@ -4,8 +4,7 @@ from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
 import pandas
 from attrs import define, field, frozen
 
-from gooddata_sdk import BareExecutionResponse, ExecutionResult
-from gooddata_sdk.compute.model.execution import ResultSize
+from gooddata_sdk import BareExecutionResponse, ExecutionResult, ResultCacheMetadata, ResultSizeDimensions
 
 _DEFAULT_PAGE_SIZE = 100
 _DataHeaders = List[List[Any]]
@@ -125,19 +124,52 @@ class _AccumulatedData:
         )
 
 
-def _extract_all_result_data(
-    response: BareExecutionResponse,
-    result_size_limits: ResultSize,
+@define
+class DataFrameMetadata:
+    # row line location where total header is located per index header column
+    # example:
+    # Category  | Country | Budget
+    # Car       | Arizona |    100
+    #           | Texas   |     50
+    #           | SUM     |    150
+    # Train     | Arizona |    200    =>    [[7],[3,6]]
+    #           | Texas   |    100
+    #           | AVG     |    150
+    # SUM       |         |    450
+    row_totals_indexes: List[List[int]]
+
+    execution_response: BareExecutionResponse
+
+    @classmethod
+    def from_data(
+        cls,
+        headers: Tuple[_DataHeaders, Optional[_DataHeaders]],
+        execution_response: BareExecutionResponse,
+    ) -> "DataFrameMetadata":
+        row_totals_indexes = [
+            [idx for idx, hdr in enumerate(dim) if hdr is not None and "totalHeader" in hdr] for dim in headers[0]
+        ]
+        return cls(
+            row_totals_indexes=row_totals_indexes,
+            execution_response=execution_response,
+        )
+
+
+def _read_complete_execution_result(
+    execution_response: BareExecutionResponse,
+    result_cache_metadata: ResultCacheMetadata,
+    result_size_dimensions_limits: ResultSizeDimensions,
+    result_size_bytes_limit: Optional[int] = None,
     page_size: int = _DEFAULT_PAGE_SIZE,
 ) -> _DataWithHeaders:
     """
     Extracts all data and headers for an execution result. This does page around the execution result to extract
     everything from the paged API.
 
-    :param response: execution response to work with;
+    :param execution_response: execution response to work with;
     :return:
     """
-    num_dims = len(response.dimensions)
+    num_dims = len(execution_response.dimensions)
     offset = [0] * num_dims
     limit = [page_size] * num_dims
     acc = _AccumulatedData()
@@ -149,14 +181,16 @@ def _extract_all_result_data(
         #
         # if one-dimensional result, it pages over an array of data
         # if two-dimensional result, it pages over table rows
-        result = response.read_result(offset=offset, limit=limit)
+        result = execution_response.read_result(offset=offset, limit=limit)
 
         if not result_size_limits_checked:
-            result.check_size_limits(result_size_limits)
+            result.check_dimensions_size_limits(result_size_dimensions_limits)
+            result_cache_metadata.check_bytes_size_limit(result_size_bytes_limit)
+            result_size_limits_checked = True
 
         acc.accumulate_data(from_result=result)
         acc.accumulate_headers(from_result=result, from_dim=0)
-        acc.accumulate_grand_totals(from_result=result, paging_dim=0, response=response)
+        acc.accumulate_grand_totals(from_result=result, paging_dim=0, response=execution_response)
 
         if num_dims > 1:
             # when result is two-dimensional make sure to read the column headers and column totals
@@ -173,12 +207,12 @@ def _extract_all_result_data(
                 # page 'to the right' to get data from all columns, extend existing rows with that data
                 offset = [offset[0], result.next_page_start(dim=1)]
                 while True:
-                    result = response.read_result(offset=offset, limit=limit)
+                    result = execution_response.read_result(offset=offset, limit=limit)
                     acc.extend_existing_row_data(from_result=result)
 
                     if load_headers_and_totals:
                         acc.accumulate_headers(from_result=result, from_dim=1)
-                        acc.accumulate_grand_totals(from_result=result, paging_dim=1, response=response)
+                        acc.accumulate_grand_totals(from_result=result, paging_dim=1, response=execution_response)
 
                     if result.is_complete(dim=1):
                         break
@@ -194,14 +228,19 @@ def _extract_all_result_data(
 
 
 def _create_header_mapper(
-    response: BareExecutionResponse, dim: int, label_overrides: LabelOverrides = None
-) -> Callable[[Any, Optional[int]], str]:
+    response: BareExecutionResponse,
+    dim: int,
+    label_overrides: Optional[LabelOverrides] = None,
+    use_local_ids_in_headers: bool = False,
+) -> Callable[[Any, Optional[int]], Optional[str]]:
     """
     Prepares header mapper function which is doing header structures translations into appropriate label used
     in a dataframe
     :param response: Response structure to gather dimension header details
     :param dim: dimension id
     :param label_overrides: label overrides
+    :param use_local_ids_in_headers: Use local identifiers of header attributes and metrics. Optional.
+
     :return: Mapper function
     """
     if label_overrides is None:
@@ -211,8 +250,8 @@ def _create_header_mapper(
     attribute_labels = label_overrides.get("labels", {})
     measure_labels = label_overrides.get("metrics", {})
 
-    def _mapper(header: Any, header_idx: Optional[int]) -> str:
-        label = ""
+    def _mapper(header: Any, header_idx: Optional[int]) -> Optional[str]:
+        label = None
         if header is None:
             pass
         elif "attributeHeader" in header:
@@ -220,20 +259,26 @@ def _create_header_mapper(
                 label = header["attributeHeader"]["labelValue"]
             elif "labelName" in header["attributeHeader"]:
                 attr_local_id = header["attributeHeader"]["localIdentifier"]
-                if attr_local_id in attribute_labels:
-                    label = attribute_labels[attr_local_id]["title"]
+                if use_local_ids_in_headers:
+                    label = attr_local_id
                 else:
-                    label = header["attributeHeader"]["labelName"]
+                    if attr_local_id in attribute_labels:
+                        label = attribute_labels[attr_local_id]["title"]
+                    else:
+                        label = header["attributeHeader"]["labelName"]
         elif "measureHeader" in header and header_idx is not None:
             measure_idx = header["measureHeader"]["measureIndex"]
             measure_descriptor = dim_descriptor["headers"][header_idx]["measureGroupHeaders"][measure_idx]
 
-            if measure_descriptor["localIdentifier"] in measure_labels:
-                label = measure_labels[measure_descriptor["localIdentifier"]]["title"]
-            elif "name" in measure_descriptor:
-                label = measure_descriptor["name"]
-            else:
+            if use_local_ids_in_headers:
                 label = measure_descriptor["localIdentifier"]
+            else:
+                if measure_descriptor["localIdentifier"] in measure_labels:
+                    label = measure_labels[measure_descriptor["localIdentifier"]]["title"]
+                elif "name" in measure_descriptor:
+                    label = measure_descriptor["name"]
+                else:
+                    label = measure_descriptor["localIdentifier"]
         elif "totalHeader" in header:
             label = header["totalHeader"]["function"]
         return label
@@ -246,11 +291,17 @@ def _headers_to_index(
     headers: Tuple[_DataHeaders, Optional[_DataHeaders]],
     response: BareExecutionResponse,
     label_overrides: LabelOverrides,
+    use_local_ids_in_headers: bool = False,
 ) -> Optional[pandas.Index]:
     if len(response.dimensions) <= dim_idx or not len(response.dimensions[dim_idx]["headers"]):
         return None
 
-    mapper = _create_header_mapper(response=response, dim=dim_idx, label_overrides=label_overrides)
+    mapper = _create_header_mapper(
+        response=response,
+        dim=dim_idx,
+        label_overrides=label_overrides,
+        use_local_ids_in_headers=use_local_ids_in_headers,
+    )
 
     return pandas.MultiIndex.from_arrays(
         [
@@ -300,11 +351,14 @@ def _merge_grand_total_headers_into_headers(extract: _DataWithHeaders) -> Tuple[
     return headers
 
 
-def convert_result_to_dataframe(
-    response: BareExecutionResponse,
+def convert_execution_response_to_dataframe(
+    execution_response: BareExecutionResponse,
+    result_cache_metadata: ResultCacheMetadata,
     label_overrides: LabelOverrides,
-    result_size_limits: ResultSize,
-) -> pandas.DataFrame:
+    result_size_dimensions_limits: ResultSizeDimensions,
+    result_size_bytes_limit: Optional[int] = None,
+    use_local_ids_in_headers: bool = False,
+) -> Tuple[pandas.DataFrame, DataFrameMetadata]:
     """
     Converts execution result to a pandas dataframe, maintaining the dimensionality of the result.
 
@@ -312,15 +366,34 @@ def convert_result_to_dataframe(
     expects that the execution _response_.
 
     :param label_overrides: label overrides
-    :param response: execution response through which the result can be read and converted to a dataframe
+    :param execution_response: execution response through which the result can be read and converted to a dataframe
     :return: a new dataframe
     """
-    extract = _extract_all_result_data(response=response, result_size_limits=result_size_limits)
+    extract = _read_complete_execution_result(
+        execution_response=execution_response,
+        result_cache_metadata=result_cache_metadata,
+        result_size_dimensions_limits=result_size_dimensions_limits,
+        result_size_bytes_limit=result_size_bytes_limit,
+    )
     full_data = _merge_grand_totals_into_data(extract)
     full_headers = _merge_grand_total_headers_into_headers(extract)
 
-    return pandas.DataFrame(
+    df = pandas.DataFrame(
         data=full_data,
-        index=_headers_to_index(dim_idx=0, headers=full_headers, response=response, label_overrides=label_overrides),
-        columns=_headers_to_index(dim_idx=1, headers=full_headers, response=response, label_overrides=label_overrides),
+        index=_headers_to_index(
+            dim_idx=0,
+            headers=full_headers,
+            response=execution_response,
+            label_overrides=label_overrides,
+            use_local_ids_in_headers=use_local_ids_in_headers,
+        ),
+        columns=_headers_to_index(
+            dim_idx=1,
+            headers=full_headers,
+            response=execution_response,
+            label_overrides=label_overrides,
+            use_local_ids_in_headers=use_local_ids_in_headers,
+        ),
     )
+
+    return df, DataFrameMetadata.from_data(headers=full_headers, execution_response=execution_response)
